/******************************************************************************
 main.c for ATTiny2313a

 This code is copyright (c) 2011 fredboboss.free.fr
 It is licensed for use under the GNU General Public License :

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

 It's nearly all work from fredboboss with adaptation to WS1600 and TX22 transmitter
 and make file for WINAVR2010 by (c) Gtz Romahn (gwr), Febr 2012
 
 This program decodes the SDO output of the WS1600 weather station receiver
 module.

 This allows to receive data transmitted by the associated
 TX22 wireless sensor.
 Data - organized in nibbles - are structured as follows (exammple with blanks added for clarity):

 a 5a 5 0 628 1 033 2 000 3 e00 4 000 bd

 data always start with "a"
 from next 1.5 nibbles (here 5a) the 6 msb are identifier of transmitter,
 bit 1 indicates acquisition/synchronizing phase (so 5a >> 58 thereafter) 
 bit 0 will be 1 in case of error (e.g. no wind sensor 5a >> 5b)
 next nibble (here 5) is count of quartets to betransmitted
 up to 5 quartets of data follow
 each quartet starts with type indicator (here 0,1,2,3,4)
 0: temperature, 3 nibbles bcd coded tenth of c plus 400 (here 628-400 = 22.8C)
 1: humidity, 3 nibbles bcd coded (here 33 %rH), meaning of 1st nibble still unclear
 2: rain, 3 nibbles, counter of contact closures
 3: wind, first nibble direction of wind vane (multiply by 22.5 to obtain degrees,
    here 0xe*22.5 = 315 degrees)
	next two nibbles wind speed in m per sec (i.e. no more than 255 m/s; 9th bit still not found)
 4: gust, speed in m per sec (yes, TX23 sensor does measure gusts and data are transmitted
    but not displayed by WS1600), number of significant nibbles still unclear
 next two bytes (here bd) are crc.
 During acquisition/synchronizing phase (abt. 5 hours) all 5 quartets are sent, see examplke above. Thereafter
 data strings contain only a few ( 1 up ton 3) quartets, so data strings are not! always of
 equal length.
 

 pin mapping is as follows :

 ATTiny2313a  / Wireless receiver module
 PD3 (INT1)     pin6 (SCK)
 PD5            pin8 (nFSEL)
 PD6            pin7 (SDO)
 
 PB4 : debug LED

 PD0 (USART RXD) => connected with serial TXD, not used
 PD1 (USART TXD) => connect to serial RXD of PC

******************************************************************************/
#define OUT_RAW		0
#define OUT_ASCII	0
#define OUT_ELV		0
#define OUT_WMR		1




#define F_CPU 8000000UL
	
//Baudrate of serial port
#if OUT_WMR
#define UART_BAUDRATE 9600
#else
#define UART_BAUDRATE 19200
#endif

#include <avr/io.h>
#include <avr/interrupt.h>  /* for sei() */
#include <avr/sleep.h>
#include <util/delay.h>

#define DEBUG_LED(reg, bit) reg |= (1 << bit)
#define DEBUG_LED_ON(port, bit)  port |= (1 << bit)
#define DEBUG_LED_OFF(port, bit) port &= ~(1 << bit)
#define DEBUG_LED_TOGGLE(port, bit) port ^= (1 << bit)



#define BIT_COUNT_MAX 104		//was 40
volatile uint8_t bit_count;
volatile uint8_t quartet_count;
volatile uint8_t quartet;
static uint8_t quartets[BIT_COUNT_MAX / 4];

#define CRC_POLY 0x31
volatile uint8_t crc;

/* timer expiry flag */
volatile uint8_t timer_ouaf;

void timer_init(void)
{
    /****************************************************************************/
    /* TIMER configuration */
    TCCR1B |= (1 << WGM12); // Configure timer 1 for CTC mode (compare with A)
    TCCR1B |= ((1 << CS10) | (1 << CS12)); // Set up timer at Fcpu/ 1024

    //OCR1A = 7812; // Set CTC compare value to 1Hz at 8MHz AVR clock, with a prescaler of 1024
//    OCR1A = 78; /* 10 ms */
    OCR1A = 39; /* 5 ms ,  reducing this value did it!! gwr* 8 MHz clock */
}

void timer_start(void)
{
    TIMSK |= (1 << OCIE1A); // Enable CTC interrupt for TIMER1

    timer_ouaf = 0;
}

void timer_reset(void)
{
    /* Reset timer value */
    TCNT1 = 0;
}

/******************************************************************************
   Function    : ISR(TIMER1_COMPA_vect)
   Parameters  : 
   Return      : 
   Description : timeout ISR :
   each time a bit is received the timer is reset (restarted from 0), 
   it it expires this means no bit has been received in 10 ms.

******************************************************************************/
ISR(TIMER1_COMPA_vect)
{
    /* set timer expiry flag */
    timer_ouaf = 1;
}


/******************************************************************************
   Function    : ioinit()
   Parameters  : 
   Return      : 
   Description : 

******************************************************************************/
static void ioinit(void)
{
    /****************************************************************************/
    /* Set Pin 6 (PD3) as the INT source (INT1) */
    PCMSK |= (1 << PIND3);

    /* interrupt on INT1 pin falling edge / FINT    */
//    MCUCR |= (1 << ISC11); // ISC10 = 0


    /* interrupt on INT1 pin rising edge / SCK */
    MCUCR |= (1 << ISC11) | (1 << ISC10);

    /* turn on INT1 interrupt */
    GIMSK |= (1 << INT1);

    /****************************************************************************/
    /* pin mapping :
       - PD3 as input/INT1
       - PD5 as input : nFSEL
       - PD6 as input : SDO
    */
    DDRD &= ~(1 << 3);
    DDRD &= ~(1 << 5);
    DDRD &= ~(1 << 6);

    /****************************************************************************/
    /* debug LED on PB4 set as output */
    DEBUG_LED(DDRB, 4);
}

/******************************************************************************
   Function    : uart_init()
   Parameters  :
   Return      :
   Description : init USART  with 8N1 / and UART_BAUDRATE

******************************************************************************/


void uart_init(void)
{
    /* Set the baud rate */
#if F_CPU < 2000000UL && defined(U2X)
    UCSRA = _BV(U2X);             /* improve baud rate error by using 2x clk */
    UBRRL = (F_CPU / (8UL * UART_BAUDRATE)) - 1;
#else
    UBRRL = (F_CPU / (16UL * UART_BAUDRATE)) - 1;
#endif

    /* Enable UART receiver and transmitter */
    UCSRB = (1 << RXEN) | (1 << TXEN);

    /* set to 8 data bits, 1 stop bit */
#if OUT_ELV
    UCSRC = (1 << UCSZ1) | (1 << UCSZ0) | (1 << USBS) | (1 << UPM1) | (1 << UPM0); //8;2;odd
#else
    UCSRC = (1 << UCSZ1) | (1 << UCSZ0);	// original 8,1,no
#endif
}

void uart_putchar(char c)
{

    if (c == '\n')
    {
        uart_putchar('\r');
    }

    //loop_until_bit_is_set(UCSRA, UDRE);
    /* Wait for empty transmit buffer */
    while (!(UCSRA & (1 << UDRE)));

    /* transmit */
    UDR = c;
}

void uart_putbyte(char c)
{
    //loop_until_bit_is_set(UCSRA, UDRE);
    /* Wait for empty transmit buffer */
    while (!(UCSRA & (1 << UDRE)));

    /* transmit */
    UDR = c;
}

/* put 4bits as a hex quartet / ASCII */
void uart_puthex(uint8_t q)
{
    /* hexa */
    if (q < 10)
    {
        uart_putchar('0' + q);
    }
    else
    {
        uart_putchar('a' + (q - 10));
    }
}


/* output 3digts from integer with trailing ' ' (blank char) / ASCII */
void uart_putint(uint16_t q)
{
		if ( q >=1000 ) {			// gusts may become up to 183.6 kmh
		uart_puthex(q/1000);
		q = q%1000;
		}
		uart_puthex(q/100);
		uart_puthex((q/10)%10);
		uart_puthex(q%10);
		uart_putchar(' ');
}

#if OUT_ASCII
/* put ASCII-string */
static void uart_putstr(const char* str)
{
    char* s = (char *)str;

    while (*s)
    {
        uart_putchar(*s);
        s++;
    }
}
#endif		//OUT_ASCII

/* not used
unsigned char uart_getchar(void)
{
    // Wait for incoming data 
    while (!(UCSRA & (1 << RXC)));

    // Return the data 
    return UDR;
}
*/


/******************************************************************************
   Function    : compute_crc()
   Parameters  : 
   Return      : 
   Description : bit by bit on the fly CRC computation for :
   CRC-8 = 0x31 / x8 + x5 + x4 + 1

******************************************************************************/
void compute_crc(uint8_t b)
{
    uint16_t reg;

    reg = crc << 1;
    reg |= b;

    if (reg & 0x100)
    {
        reg ^= CRC_POLY;
    }

    crc = reg & 0xFF;
}

/******************************************************************************
   Function    : SIGNAL (INT1_vect)
   Parameters  :
   Return      :
   Description : interrupt routine on PD3/SCK

   this routine reads SDO bits, compute on the fly CRC and assemble quartets
******************************************************************************/
SIGNAL (INT1_vect)
{
    /* get D port value */
    uint8_t x = PIND;

    /* PD5 = 0 ? only acquire if PD5 (nFSEL) is low */
    if (0 == ((x >> 5) & 1))
    {
        timer_reset();

        if (bit_count < BIT_COUNT_MAX)
        {
            /* PD6 ? */
            uint8_t b = (x >> 6) & 1;

            /* add data bit to current quartet */
            quartet <<= 1;
            quartet |= b;
            bit_count++;

            /* add quartet to quartet array */
            if (0 == (bit_count % 4))
            {
                quartets[quartet_count] = quartet;
                quartet_count++;

                /* reset */
                quartet = 0;
            }

            /* compute CRC on the fly */
            compute_crc(b);
        }
    }
}

#if OUT_RAW
/******************************************************************************
   Function    : display_message(void), raw nibbles useful for diagnosing
   Parameters  : 
   Return      : 
   Description : display stored quartets and CRC over UART

******************************************************************************/
void display_message(void)
{
    uint8_t i;
    for (i = 0; i < quartet_count; i++)
    {
        uart_puthex(quartets[i]);
    }
    uart_putchar(' ');
    /* display CRC */
    uart_puthex(crc >> 4);
    uart_puthex(crc & 0xF);
    
    uart_putchar('\n');
}
#endif 		// OUT_RAW

#if OUT_ASCII
/******************************************************************************
   Function    : display_values(void)
   Parameters  : 
   Return      : 
   Description : display decoded data over UART

******************************************************************************/
void display_values(void)
{

int temp;
uint8_t i;

if ( (quartets[0] != 0xa) | (crc != 0) ) return;	// if bad data no display

for ( i = 1; i < quartets[3]+1; i++) {
	
	switch( quartets[i*4] ) {
	case 0:
		uart_putstr("Temp ");
//		uart_putchar('T');
		temp = (quartets[i*4+1]*100 + quartets[i*4+2]*10 + quartets[i*4+3])-400;
		if (temp < 0) {
			uart_putchar('-');
			temp = -temp;
		} else uart_putchar(' ');
		uart_putint(temp);
	break;
	case 1:
		uart_putstr("Humi ");
//		uart_putchar('H');
		uart_puthex(quartets[i*4+2]);
		uart_puthex(quartets[i*4+3]);
		uart_putchar(' ');
	break;
	case 2:
		uart_putstr("Rain ");
//		uart_putchar('R');
		temp = ( quartets[i*4+3] + (quartets[i*4+2]<<4) + ((int)quartets[i*4+1]<<8) )*5;
		uart_putint(temp);
	break;
	case 3:
		uart_putstr("Wind ");
//		uart_putchar('W');
		temp = ( quartets[i*4+3] + (quartets[i*4+2]<<4) )*36/10;
		uart_putint(temp);
		uart_putstr(" Dir ");
//		uart_putchar('D');
		temp = (quartets[i*4+1]*225/10);
		uart_putint(temp);
	break; 
	case 4:
		uart_putstr("Gust ");
//		uart_putchar('B');
		temp = ( quartets[i*4+3] + (quartets[i*4+2]<<4) + ((int)quartets[i*4+1]<<8) )*36/10;
		uart_putint(temp);
	break; 
	}    
}
    uart_putchar('\n');
}
#endif		// OUT_ASCII

char humi;
int dire;
int temp;
int rain;
int wind;
int gust;

/******************************************************************************
   Function    : display_values(void)
   Parameters  : 
   Return      : 
   Description : output decoded data over UART in ELV strings

******************************************************************************/
#if OUT_ELV
char frame[6];

void out_frame(void)
{
uint8_t i;
	uart_putchar(0x02);					// STX
	for ( i = 0; i < 6; i++) {
		uart_putchar(frame[i] | 0x80);
		frame[i] = 0;
	}
	uart_putchar(0x03);					//ETX
}

void display_values(void)
{
uint8_t i;

if ( (quartets[0] != 10) | (crc != 0) ) return;	// if bad data no display

for ( i = 1; i < quartets[3]+1; i++) {
	
	switch( quartets[i*4] ) {
	case 0:
		temp = (quartets[i*4+1]*100 + quartets[i*4+2]*10 + quartets[i*4+3])-400;
		frame[0] = 0x18;					// t/h
		frame[1] = (temp >> 7) & 0x7f;
		frame[2] = temp & 0x7f;
		frame[3] = humi & 0x7f;
		out_frame();
	break;
	case 1:
		humi = quartets[i*4+2]*10 + quartets[i*4+3];
	break;
	case 2:
		rain = ( quartets[i*4+3] + (quartets[i*4+2]<<4) + ((int)quartets[i*4+1]<<8) );
		frame[0] = 0x28;					// rain
		frame[1] = (rain >> 7) & 0x1f;
		frame[2] = rain & 0x7f;
		out_frame();
	break;
	case 3:
		if  (quartets[2] & 0x01) break;	// if bad wind data no display
		wind = ( quartets[i*4+3] + (quartets[i*4+2]<<4) )*36/10;
		if ( wind == 914 ) wind =14;		// erratic value, generated by ?? gwr
		dire = (quartets[i*4+1]*225/10);
		frame[0] = 0x38;					// wind
		frame[1] = (wind >> 7) & 0x7f;
		frame[2] = wind & 0x7f;
//		frame[3] = 0x00;
		frame[4] = (dire >> 7) & 0x7f;
		frame[5] = dire & 0x7f;
		out_frame();
	break; 
	case 4:
		if  (quartets[2] & 0x01) break;	// if bad wind data no display
		gust = ( quartets[i*4+3] + (quartets[i*4+2]<<4) + ((int)quartets[i*4+1]<<8) )*36/10;
		frame[0] = 0x4a;					// wind/gust
		frame[1] = (gust >> 7) & 0x7f;
		frame[2] = gust & 0x7f;
//		frame[3] = 0x0;
		frame[4] = (dire >> 7) & 0x7f;
		frame[5] = dire & 0x7f;
		out_frame();
	break; 
	}    
}
//    uart_putchar('\n');
}
#endif		// OUT_ELV


/******************************************************************************
   Function    : display_values(void)
   Parameters  : 
   Return      : 
   Description : output decoded data over UART in WMR928 strings

******************************************************************************/
#if OUT_WMR
char frame[15];
uint8_t sign;
uint8_t size;
uint16_t wmrcrc;

void out_frame(void)
{
	uint8_t i;
	wmrcrc = 0;
	for ( i = 0; i < size; i++) {
		uart_putbyte(frame[i]);
		wmrcrc = wmrcrc + frame[i];
	}
		uart_putbyte(wmrcrc &0xff);
		_delay_ms (10);
}

	uint16_t minu=0;
	uint8_t hour = 0;

void display_values(void)
{
	uint8_t i;

if ( (quartets[0] != 10) | (crc != 0) ) return;	// if bad data no display

		size = 4;
		minu++;
		frame[0] = 0xff;
		frame[1] = 0xff;
		frame[2] = 0x0e;		//minute = received packet ticks every 4.5 secs
		frame[3] = ((minu/10<<4) + minu%10) & 0x7f;
		out_frame();

		if (minu == 800) {		// 800 ticks abt. 1 hour
		minu= 0;
		hour++;
		}

		size = 8;
		frame[2] = 0x0f;		// hour
		frame[3] = 0;
		frame[4] = (hour/10<<4) + hour%10;
		frame[5] = 0x01;
		frame[6] = 0x01;
		frame[7] = 0x12;
		out_frame();
	

// send some dummy inside values otherwise "cumulus" will not log any data
		size = 12;
		frame[2] = 0x05;
		frame[4] = 0x23;
		frame[5] = 0x01;
		frame[6] = humi;
		frame[7] = 0;
		frame[8] = 205;
		frame[9] = 0x0c;
		frame[10] = 0;
		frame[11] = 0;
		out_frame();

for ( i = 1; i < quartets[3]+1; i++) {

	switch( quartets[i*4] ) {
	case 0:
		sign = 0;
		temp = (quartets[i*4+1]*100 + quartets[i*4+2]*10 + quartets[i*4+3])-400;
		if (temp <0) {
			sign = 0x80;
			temp = -temp; }
		size = 8;
		frame[2] = 0x03;
//		frame[3] = 0x01;
		frame[4] = (temp%100/10<<4) + temp%10;
		frame[5] = (temp/100) + sign;
		frame[6] = humi;
		frame[7] = 0x00;
		out_frame();
	break;
	case 1:
		humi = (quartets[i*4+2]<<4) + quartets[i*4+3];
	break;
	case 2:
		rain = ( quartets[i*4+3] + (quartets[i*4+2]<<4) + ((int)quartets[i*4+1]<<8) )>>1;
		size = 15;
		frame[2] = 0x01;
//		frame[3] = 0x00;
		frame[4] = 0x00;
		frame[5] = 0x60;
		frame[6] = (rain%100/10<<4) + rain%10;
		frame[7] = (rain/1000<<4) + rain%1000/100;
		frame[8] = frame[6];
		frame[9] = frame[7];
		frame[10] = 0x45;
		frame[11] = 0x17;
		frame[12] = 0x01;
		frame[13] = 0x03;
		frame[14] = 0x12;
		out_frame();
	break;
	case 3:
		if  (quartets[2] & 0x01) break;	// if bad wind data no display
		wind = ( quartets[i*4+3] + (quartets[i*4+2]<<4) );
//		if ( wind == 914 ) wind =14;		// erratic value, generated by ?? gwr
		dire = (quartets[i*4+1]*225/10);
		size = 10;
		frame[2] = 0x00;
//		frame[3] = 0x00;
		frame[4] = (dire%100/10<<4) + dire%10;
		frame[5] = (dire/100) + (gust%10<<4);
		frame[6] = (gust/100<<4) + gust%100/10;
		frame[7] = (wind%100/10<<4) + wind%10;
		frame[8] = wind/100;
		frame[9] = 0;
		out_frame();
	break; 
	case 4:
		if (quartets[2] & 0x01) break;	// if bad wind data no display
		gust = ( quartets[i*4+3] + (quartets[i*4+2]<<4) + ((int)quartets[i*4+1]<<8) );
/*
		size = 10;
		frame[2] = 0x00;
		frame[4] = (dire%100/10<<4) + dire%10;
		frame[5] = (dire/100) + (gust%10<<4);
		frame[6] = (gust/100<<4) + gust%100/10;
		frame[7] = (wind%100/10<<4) + wind%10;
		frame[8] = wind/100;
		frame[9] = 0;
		out_frame();
*/
	break; 
	}    

}
}
#endif		// OUT_WMR


/******************************************************************************
   Function    : main()

******************************************************************************/
int main(void)
{
    /* call our I/O init */
    ioinit();

    timer_init();
    timer_start();

    uart_init();

//	uart_putstr("test");

    /* reset counts & CRC */
    bit_count = 0;
    quartet_count = 0;
    quartet = 0;
    crc = 0;

    sei();

    for (;;)
    {
        set_sleep_mode(SLEEP_MODE_IDLE);
        sleep_mode();                   // in den Schlafmodus wechseln

       // hier wachen wir wieder auf
         if (timer_ouaf)
        {
            if (quartet_count)
            {
                /* we will have a short flash every 4seconds */
                DEBUG_LED_ON(PORTB, 4);

				if (quartet_count >= 10) {
				#if OUT_RAW
				display_message();			// raw data
				#endif
				#if OUT_ASCII | OUT_ELV | OUT_WMR
                display_values();			// decoded data
				#endif
				}

                /* reset bitcount */
                bit_count = 0;
                quartet_count = 0;
                quartet = 0;
                crc = 0;

                DEBUG_LED_OFF(PORTB, 4);
            }

            /* ack expiry */
            timer_ouaf = 0;
        }
    }

    return 0;
}
